#!/usr/bin/perl 
#
# sherpa v0.1.8 - a host security tool
#
# $Id: sherpa.pl,v 1.2 2004/11/04 07:35:52 rcrelia Exp $
#
# Copyright (C) 2000-2004 Rick Crelia <rcrelia@users.sourceforge.net>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#######################################################################
#
#**********************************************************************
# Local configuration options...modify as needed
#**********************************************************************

*sherpa_home = *FindBin::RealBin;   # Alias sherpa_home to where sherpa.pl lives

$config_loc = "/etc";               # Location of common system config
                                    # files

$log_dir = "/var/log";              # Location of system log files

$perms_list = "$sherpa_home/perms.lst";    # Location of file permission list

# This is the directory where the HTML version of sherpa's
# output will go. Used only when $html is set
$sherpa_webroot = "/var/www/htdocs/sherpa";

$sherpa_log = "$log_dir/sherpa";           # log file location

# What services should sherpa evaluate?
#
# These are potentially risky services, but often come with Unix/Linux
# in an active state. sherpa will check for their usage and report it

@badsvcs = (
    "echo",      "discard", "daytime",     "ftp",
    "smtp",      "exec",    "chargen",     "ntalk",
    "dtalk",     "talk",    "telnet",      "tftp",
    "bootps",    "finger",  "systat",      "netstat",
    "linuxconf", "shell",   "login",       "uucp",
    "pop3",      "auth",    "netbios-ssn", "netbios-ns",
    "swat",      "rstatd",  "rusersd",     "walld"
);

# These are used in report generation only
$today  = `date`; $ftoday = `date +%Y%m%d`;
chop($ftoday);

$host = $ENV{"HOSTNAME"};

# list of directories to search for world writable, SGID, SUID,
# .rhosts and orphaned files & sub-dirs. Note: fs boundaries
# ARE respected

$dirs2find = "$sherpa_home/find.lst";

#**********************************************************************
# End of local configuration options - do not modify below this point
# unless you know what you are doing...
#**********************************************************************

$ver = "0.1.8";

# hashs for use by check_orphans
%uid, %user, %gid, %group;

my $hr_str = '-';
foreach my $count (1..78)
{
  $hr .= $hr_str;
} 
$hr .= "\n\n";

use File::Find;
use Getopt::Std;
use FindBin;

Main:

if ( $< != 0 ) {
    print STDERR "Sorry, you must be root to run this program\n";
    exit 1;
}

&parseopts();

&setup();

&check_sugids() if $chksuids;

&check_configs() if $chkconfigs;

&check_privs() if $chkprivs;

&check_orphans() if $chkorphans ;

&tidy();

close;

exit 0;

#######################################################################

# Subroutines follow

#######################################################################

sub tidy {
    $finish = localtime(time);
    $stop   = time;
    $rtime  = $stop - $begin;
    $rmin   = int( $rtime / 60 );
    $rsec   = $rtime % 60;            
    $msg    = "$hr" . "[$finish] - sherpa finished\n\n";
    &prtmsg();
    print LOG $msg if $logging;
    $msg = "\nRunning time = $rmin minutes $rsec seconds\n";
    &prtmsg();
    print LOG $msg if $logging;
    $msg = "\n-=< Results are in $sherpa_outfile >=-\n\n";
    &prtmsg() if $logging ;

    &html_finish() if $html;

    close;
}

#######################################################################

sub check_privs {

    $now = localtime(time);
    $msg = "$hr" . "[$now] - sherpa checking file and directory permissions\n";
    &prtmsg();
    &perm_chk();

    $now = localtime(time);
    $msg = "$hr" . "[$now] - sherpa checking for world writable files\n";
    &prtmsg();
    &wwrite();

    $msg = "\n";
    &prtmsg();

}

#######################################################################

sub wwrite {
    $wwcntr = 0;
    find( \&findww, @fndlist );
    if ( $wwcntr > 0 ) {
        $msg = "\n$wwcntr world-writable files found:\n\n";
        &prtmsg();

        for ( $i = 0 ; $i < $wwcntr ; $i++ ) {
            $j   = $i + 1;
            $j   = " " . $j if ( $j < 10 );
            $msg = "$j\. $wwlist[$i]\n";
            &prtmsg();
        }
    }
    else {
        $msg = "\nNo world-writable files found. Good!\n\n";
        &prtmsg();
    }
}

#######################################################################

sub findww {

    my ( $dev, $ino, $mode, $nlink, $uid, $gid );

    if ( ( ( $dev, $ino, $mode, $nlink, $uid, $gid ) = lstat($_) )
        && !( $File::Find::prune |= ( $dev != $File::Find::topdev ) ) )
    {

        if ( ( ( $mode & 02 ) == 02 ) && -f && !-l ) {
            push( @wwlist, $File::Find::name );
            $wwcntr++;
        }
    }
}

#######################################################################

sub perm_chk {
    open PERMS, "$perms_list" || die "Cannot open permissions list!\n";
  PERMCHK:
    while ( $line = <PERMS> ) {
        if ( $line =~ /^\#/ || $line =~ /^$/ ) {
            next PERMCHK;
        }
        else {
            chop($line);
            ( $target, $perms, $uname, $gname ) = split( /,/, $line );
            $uid = getpwnam $uname;
            $gid = getgrnam $gname;
            if ( !-e $target ) {
                $msg = "$target not found!\n";
                &prtmsg();
                next PERMCHK;
            }

            &get_fileinfo($target);

            if ( $found_perms != $perms ) {
                $msg =
                  "$target has wrong permissions =  $found_perms vs. $perms\n";
                &prtmsg();
                if ( $fix == 1 ) {
                    $msg = "Changing permissions for $target to $perms\n";
                    &prtmsg();
                    chmod oct($perms), $target || do {
                        $msg =
                          "Can't change permissions for $target to $perms\n";
                        &prtmsg();
                      }
                }
            }
            else {
                $msg = "$target has correct permissions = $found_perms\n";
                &prtmsg() if $debug ;
            }

            if ( $found_uid != $uid || $found_gid != $gid ) {
                $msg =
                    "$target has wrong ownership - UID = $found_uname "
                  . "GID = $found_gname\n";
                &prtmsg();
                if ( $fix == 1 ) {
                    $msg =
                        "Changing ownership for $target to UID = $uname and "
                      . "GID = $gname\n";
                    &prtmsg();
                    chown $uid, $gid, $target || do {
                        $msg =
                            "Can't change owner for $target to UID "
                          . "$uname GID $gname\n";
                        &prtmsg();
                      }
                }
            }
            else {
                $msg =
                    "$target has correct ownership - UID = $uname "
                  . "and GID = $gname\n";
                &prtmsg() if $debug ;
            }

        }

    }
    close(PERMS);
    $msg = "\n";
    &prtmsg();
}

#######################################################################

sub check_configs {

    $now = localtime(time);
    $msg = "$hr" . "[$now] - sherpa checking config files in $config_loc\n";
    &prtmsg();

    ## Network security checks
    $now = localtime(time);
    $msg = "[$now] - \tchecking network security\n";
    &prtmsg();

    &check_hosts_equiv();

    &check_rhosts();

    &check_inetd();

    &check_pass();

    $msg = "\n";
    &prtmsg();

}

#######################################################################

sub check_inetd {
    $msg = "\n";
    &prtmsg();
    $target = "$config_loc/inetd.conf";
    open( INETD, $target ) || return;
  SVCCHK:
    while ( $line = <INETD> ) {
        if ( $line =~ /^\#/ || $line =~ /^$/ ) {
            next SVCCHK;    #skip..service has been commented out
        }
        else {
            chop($line);
            ( $svc, $stype, $sprot, $option, $user, $d1, $d2 ) = split /\s+/,
              $line, 7;
            if ( $svc ne $lastsvc ) {
                push( @services, $svc );    # avoid duplicates
                $svccntr++;
                if ( $d1 eq "/usr/sbin/tcpd" ) {
                    push( @wrapped, $svc );
                    $wrapcntr++;
                }
            }
            $lastsvc = $svc;
        }
    }
    close(INETD);
    $msg = "Number of active services: $svccntr\n";
    &prtmsg();
    $msg = "Number of wrappered active services: $wrapcntr\n";
    &prtmsg();
    $msg = "Active services found (in $config_loc/inetd.conf):\n\n";
    &prtmsg();

    for ( $i = 0 ; $i < $svccntr ; $i++ ) {
        $j   = $i + 1;
        $j   = " " . $j if ( $j < 10 );
        $msg = "\t$j\. $services[$i]\n";
        &prtmsg();

        #check for problem svcs
        for my $badsvc (@badsvcs) {

            # print $services[$i], ' ', $badsvc, "\n";
            if ( $services[$i] =~ /$badsvc/ ) {
                push( @riskysvc, $badsvc );
                $badsvccntr++;
            }
        }
    }

    if ( $badsvccntr > 0 ) {
        $msg = "\n";
        &prtmsg();
        for ( $k = 0 ; $k < $badsvccntr ; $k++ ) {
            $msg = "Warning - $riskysvc[$k] can be risky to use\n";
            &prtmsg();
        }
    }

    $msg = "\n";
    &prtmsg();
    $msg = "Wrappered services found (in $config_loc/inetd.conf):\n\n";
    &prtmsg();
    for ( $i = 0 ; $i < $wrapcntr ; $i++ ) {
        $j   = $i + 1;
        $j   = " " . $j if ( $j < 10 );
        $msg = "\t$j\. $wrapped[$i]\n";
        &prtmsg();
    }

    if ( $svccntr != $wrapcntr ) {
        $msg = "\n";
        &prtmsg();
        $msg = "Warning - Not all active services are wrappered!\n";
        &prtmsg();
        $msg = "If this includes internal services (e.g., echo, daytime)\n";
        &prtmsg();
        $msg = "then it's not a huge problem. However, consider turning\n";
        &prtmsg();
        $msg = "internal services off unless you really need them.\n";
        &prtmsg();
    }
    $msg = "\n";
    &prtmsg();
}

#######################################################################

sub check_pass {
    $now = localtime(time);
    $msg = "[$now] - sherpa checking password files for shadowing\n\n";
    &prtmsg();
    open PASSWD, "/etc/passwd" || die "Cannot open /etc/passwd!\n";
    $entry = <PASSWD>;
    if ( $entry =~ /.*:x:/ ) {
        $msg = "\t User passwords are shadowed ... good!\n";
        &prtmsg();
    }
    else {
        $msg = "\t User passwords are NOT shadowed!\n";
        &prtmsg();
        if ( $fix == 1 ) {
            $msg = "\t Shadowing user passwords with pwconv...\n";
            &prtmsg();
            system "/usr/sbin/pwconv";
        }
        else {
            $msg = "Consider running pwconv to shadow your passwd file!\n";
            &prtmsg();
        }
    }

    open GROUP, "/etc/group" || die "Cannot open /etc/group!\n";
    $entry = <GROUP>;
    if ( $entry =~ /.*:x:/ ) {
        $msg = "\t Group passwords are shadowed ... good!\n";
        &prtmsg();
    }
    else {
        $msg = "\t Group passwords are NOT shadowed!\n";
        &prtmsg();
        if ( $fix == 1 ) {
            $msg = "Shadowing group passwords with grpconv...\n";
            &prtmsg();
            system "/usr/sbin/grpconv";
        }
        else {
            $msg = "Consider running grpconv to shadow your group file!\n";
            &prtmsg();
        }
    }
}

#######################################################################

sub check_rhosts {
    find( \&find_rhosts, @fndlist );
    if ( $rhostcntr > 0 ) {
        @sorted_rhosts = sort @rhosts;
        @rhosts        = @sorted_rhosts;
        $msg           = "\n** WARNING: Found $rhostcntr .rhosts files\n\n";
        &prtmsg();
        for ( $i = 0 ; $i < $rhostcntr ; $i++ ) {
            &get_fileinfo( $rhosts[$i] );
            $j   = $i + 1;
            $msg = "$j\. $rhosts[$i] - Perms = $found_perms\n";
            &prtmsg();
        }
    }
    else {
        $now = localtime(time);
        $msg = "[$now] - \t   No .rhosts files found...good!\n";
        &prtmsg();
    }
    $msg = "\n";
    &prtmsg();
}

#######################################################################

sub find_rhosts {

    my ( $dev, $ino, $mode, $nlink, $uid, $gid );

    if ( ( ( $dev, $ino, $mode, $nlink, $uid, $gid ) = lstat($_) )
        && !( $File::Find::prune |= ( $dev != $File::Find::topdev ) ) )
    {
        if ( /^\.rhosts$/ && -e && -f && !-l ) {
            push( @rhosts, $File::Find::name );
            $rhostcntr++;
        }
    }
}

#######################################################################

sub check_hosts_equiv {

    # check for hosts.equiv
    $target = "$config_loc/hosts.equiv";
    if ( -e $target ) {
        $msg =
            "\n** WARNING: $target found!\n** Do you really need to "
          . "establish this level of trust with other systems?!?\n**\n\n";
        &prtmsg();

        # check contents
        open( HE, "$target" );
        while (<HE>) {
            if (/^\+$/) {
                $msg =
                    "** Found a + by itself on a line in $target... "
                  . "not a good thing!\n\n";
                &prtmsg();
            }
            if (/^\-/) {
                $msg =
                    "** Found a - as first character in $target... "
                  . "not a good thing!\n** See CERT advisory [CA-91:12]\n\n";
                &prtmsg();
            }
        }
        close(HE);
        &get_fileinfo($target);
        $safe_perms = 600;
        if ( $found_perms != $safe_perms ) {
            $msg = "$target has incorrect permissions - $found_perms \n";
            &prtmsg();
            if ( $fix == 1 ) {
                $msg = "Changed permissions on $target to $safe_perms!\n";
                chmod oct($safe_perms), $target || do {
                    $msg = "Can't change permissions for $target to $perms\n";
                  }
                  & prtmsg();
            }
        }
        else {
            $msg = "Permissions on $target are $safe_perms ... good!\n\n";
            &prtmsg();
        }
    }
    else {
        $now = localtime(time);
        $msg = "[$now] - \t   No hosts.equiv file found...good!\n";
        &prtmsg();
    }

    $msg = "\n";
    &prtmsg();
}

#######################################################################

sub check_sugids {

    @uglist = ();
    @ulist  = ();
    @glist  = ();
    $ugcntr = 0;
    $gcntr  = 0;
    $ucntr  = 0;
    $now    = localtime(time);
    $msg = "$hr" . "[$now] - sherpa checking for SUID and SGID files & dirs\n";
    &prtmsg();
    find( \&finduids, @fndlist );

    @sorted_uglist = sort @uglist;
    @uglist        = @sorted_uglist;
    $msg           = "\n$ugcntr SUID and SGID files found:\n\n";
    &prtmsg();
    for ( $i = 0 ; $i < $ugcntr ; $i++ ) {
        $j   = $i + 1;
        $msg = "$j\. $uglist[$i] is SUID and SGID\n";
        &prtmsg();
    }

    @sorted_ulist = sort @ulist;
    @ulist        = @sorted_ulist;
    $msg          = "\n$ucntr SUID files found:\n\n";
    &prtmsg();
    for ( $i = 0 ; $i < $ucntr ; $i++ ) {
        $j   = $i + 1;
        $msg = "$j\. $ulist[$i] is SUID\n";
        &prtmsg();
    }

    @sorted_glist = sort @glist;
    @glist        = @sorted_glist;
    $msg          = "\n$gcntr SGID files found:\n\n";
    &prtmsg();
    for ( $i = 0 ; $i < $gcntr ; $i++ ) {
        $j   = $i + 1;
        $msg = "$j\. $glist[$i] is SGID\n";
        &prtmsg();
    }
    $msg = "\n";
    &prtmsg();
}

#######################################################################

sub finduids {

    my ( $dev, $ino, $mode, $nlink, $uid, $gid );

    if ( ( ( $dev, $ino, $mode, $nlink, $uid, $gid ) = lstat($_) )
        && !( $File::Find::prune |= ( $dev != $File::Find::topdev ) ) )
    {
        if ( -u && -g ) {
            push( @uglist, $File::Find::name );
            $ugcntr++;
        }
        elsif ( -u && !-g ) {
            push( @ulist, $File::Find::name );
            $ucntr++;
        }
        elsif ( -g && !-u ) {
            push( @glist, $File::Find::name );
            $gcntr++;
        }
    }

}

#######################################################################

sub check_orphans {

    @uglist = ();
    @ulist  = ();
    @glist  = ();
    $ugcntr = 0;
    $gcntr  = 0;
    $ucntr  = 0;
    $now    = localtime(time);
    $msg    = "$hr" . "[$now] - sherpa checking for orphaned files & dirs\n";
    &prtmsg();

    File::Find::find( \&findorphans, @fndlist );

    @sorted_uglist = sort @uglist;
    @uglist        = @sorted_uglist;
    $msg           = "\n$ugcntr orphaned files found no user, no group:\n\n";
    &prtmsg();
    for ( $i = 0 ; $i < $ugcntr ; $i++ ) {
        $j   = $i + 1;
        $msg = "$j\. $uglist[$i] has no owner or group\n";
        &prtmsg();
    }

    @sorted_ulist = sort @ulist;
    @ulist        = @sorted_ulist;
    $msg          = "\n$ucntr orphaned files found, no user:\n\n";
    &prtmsg();
    for ( $i = 0 ; $i < $ucntr ; $i++ ) {
        $j   = $i + 1;
        $msg = "$j\. $ulist[$i] has no owner\n";
        &prtmsg();
    }

    @sorted_glist = sort @glist;
    @glist        = @sorted_glist;
    $msg          = "\n$gcntr orphaned files found, no group:\n\n";
    &prtmsg();
    for ( $i = 0 ; $i < $gcntr ; $i++ ) {
        $j   = $i + 1;
        $msg = "$j\. $glist[$i] has no group\n";
        &prtmsg();
    }
    $msg = "\n";
    &prtmsg();
}

#######################################################################

sub findorphans {

 my ( $dev, $ino, $mode, $nlink, $uid, $gid );

 if ( ( ( $dev, $ino, $mode, $nlink, $uid, $gid ) = lstat($_) )
        && !( $File::Find::prune |= ( $dev != $File::Find::topdev ) ) )
    {

        if ( !exists $uid{$uid} && !exists $gid{$gid} ) {
            push( @uglist, $File::Find::name );
            $ugcntr++;
        }
        elsif ( !exists $uid{$uid} ) {
            push( @ulist, $File::Find::name );
            $ucntr++;
        }
        elsif ( !exists $gid{$gid} ) {
            push( @glist, $File::Find::name );
            $gcntr++;
        }
    }
}

#######################################################################

sub setup {
    open( FLIST, $dirs2find ) || die "$!: $dirs2find";
    while ( $dname = <FLIST> ) {
        chop($dname);
        push( @fndlist, $dname );
    }
    close(FLIST);

    if ( $chkorphans ) {
        while ( my ( $name, $pw, $uid ) = getpwent ) {
            $uid{$name} = $uid{$uid} = $uid;
        }
        while ( my ( $name, $pw, $gid ) = getgrent ) {
            $gid{$name} = $gid{$gid} = $gid;
        }
    }

    if ( $html ) {
        $sherpa_outfile = "$sherpa_webroot/$ftoday.html";
        open( OUT, "> $sherpa_outfile" )
          || die "Cannot create $sherpa_outfile!";
        &html_begin();
    }
    elsif ( $logging ) {
        $sherpa_outfile = "$sherpa_home/sherpa.results.txt";
        open( OUT, "> $sherpa_outfile" )
          || die "Cannot create $sherpa_outfile!";
        $msg = "sherpa $ver results - $host - $today\n\n";
        &prtmsg();
    }

    $start = localtime(time);
    $begin = time;
    $msg   = "[$start] - sherpa started\n";
    &prtmsg();
    if ( $logging ) {
        open( LOG, ">> $sherpa_log" ) || die "Cannot write to logfile!\n";
        print LOG $msg;
    }

    $msg = "\n";
    &prtmsg();

}

#######################################################################

sub get_fileinfo {
    $found_perms = sprintf( "%o", ( stat( $_[0] ) )[2] & oct(7777) );
    $found_uid   = ( stat( $_[0] ) )[4];
    $found_gid   = ( stat( $_[0] ) )[5];
    $found_perms;
    $found_uid;
    $found_gid;
    $found_uname = getpwuid $found_uid;
    $found_gname = getgrgid $found_gid;
    $found_uname;
    $found_gname;
}

#######################################################################

sub prtmsg {
    print $msg if $verbose;
    print OUT $msg if $logging;
}

#######################################################################

sub html_begin {
    print OUT <<"HTML_START";
\n\n
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">
<html>
<head>
<title>sherpa $ver results - $host - $today</title>
</head>
<body>
<h3>sherpa $ver results - $host - $today</h3>
<hr>
<p>
<pre>

HTML_START
}

#######################################################################

sub html_finish {
    print OUT <<"HTML_END";

</pre>
<p>
<hr>
</body>
</html>
HTML_END
}

#######################################################################

sub parseopts {

    getopts("vhVdflowscpr:");

    # undef = minimal output to STDOUT
    # set = verbose output to STDOUT
    $verbose = $opt_V || undef;

    # extra output undef=no <default>, set=yes
    $debug = $opt_d || undef;

    # Do you want the output to be an HTML page?
    # undef=no <default>, set=yes
    $html = $opt_w || undef;

    # Should sherpa create a logfile ?
    # undef=yes <default> set=no
    if ($opt_l) {
     $logging = undef;
    }
    else {
     $logging = 1;
    }

    ## Which checks do you want to run?
    # specified = yes in each instance

    # check for SUID and SGID files/dirs
    $chksuids = $opt_s || undef;

    # check common config files
    $chkconfigs = $opt_c || undef;

    # check file & dir permissions and ownerships
    $chkprivs = $opt_p || undef; 

    # check for orphaned files/dirs
    $chkorphans = $opt_o || undef;

    # Do you want sherpa to correct problems that it can?
    #
    # NOTE: The first time you run sherpa, leave this OFF and examine
    # the output file to get an idea of what would be changed
    # 0=no <default> 1=yes

    if ($opt_f) {
        if ( !$opt_p ) {
            print "\n    You must specify -p to use -f\n\n";
            exit 1;
        }
        $fix = 1;
    }
    else {
        $fix = 0;
    }

    if ($opt_v) {
        print "\nsherpa.pl v$ver - a host security tool\n\n";
        exit;
    }

    if ($opt_h) {
        print "\nsherpa.pl v$ver - a host security tool\n\n";
        print "Usage: sherpa.pl [-hvVdwlscpf]

	-h  print this message
	-v  print version information
       	-V  turn off verbose output
	-d  print debugging output
	-w  generate output in HTML format
       	-l  turn logging off
       	-o  check for orphaned files
       	-s  check for SUID/SGID files
       	-c  check system configuration files
	-p  check file/dir permissions in perms.lst
	-f  fix file and directory permissions [requires -p]

     \n";
        exit;
    }

    if ( !$opt_s && !$opt_c && !$opt_p && !$opt_o ) {
        print "\nsherpa.pl v$ver - a host security tool\n";
        print "\nYou did not specify any checks to perform.\n";
        print "Try -s, -c, -o, -p or -pf (or some combination thereof)\n\n";
        print "For a list of options, type sherpa.pl -h\n\n";
        exit 1;
    }

}
